// avrdude -v -pattiny13 -cusbasp -B10 -U lfuse:w:0x3a:m -U hfuse:w:0xff:m 
//             +-\/-+
//     RESET  1|    |8  Vcc
//  (D3) PB3  2|    |7  PB2 (D2) 
//  (D4) PB4  3|    |6  PB1 (D1) 
//       GND  4|    |5  PB0 (D0) 
//             +----+
// schematic file:///Volumes/Lion-Data/home/Documents/Arduino/shoe-led/schematic.pdf
// Fuses: -U lfuse:w:0x3a:m -U hfuse:w:0xff:m 
// changed hardware/attiny13/avr/platform.txt to "compiler.cpp.flags=-c -g -Os -w -fno-exceptions -ffunction-sections -fdata-sections -fno-threadsafe-statics -MMD -fshort-enums" (56 bytes saved)



#include <stdlib.h>
//#include <string.h>
#include <math.h>
#include <util/atomic.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#define __DELAY_ROUND_CLOSEST__
#include <avr/delay.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>

//#define RIGHT_SHOE
#include "shoe-led-registers.h"


// Types

typedef enum {SHOE_DOWN, SHOE_UP, SHOE_LOWERING} shoeStateType; 


// Constants

const PROGMEM uint8_t GAMMA[17] = {0, 1, 1, 1, 2, 2, 3, 4, 4, 5, 6, 8, 9, 11, 12, 14, 16}; //0..16


// Variables

volatile uint16_t _millis = 1; // Overflows once every minute (65.5 seconds) //!!!!!! can't be initialized to 0 for some reason, or millis jumps to around 1-2 seconds before overflow
volatile uint8_t leds[20] = {0,}; // brightness of each led (0..5)
//todo: 19 iso 20
volatile shoeStateType shoeState = SHOE_DOWN; 


// Function definitions

void __attribute__((always_inline)) init();
void __attribute__((always_inline)) sleep();
void __attribute__((always_inline)) splash();
void __attribute__((always_inline)) twinkle();
void setLed(uint8_t index, uint8_t value);
void delay(uint16_t ms);
uint16_t millis();
//void setAllLeds(uint8_t brightness);
//void flashLedsIndex();



int main() 
{
    init();

    shoeStateType oldShoeState = shoeState;

    while( true )
    {
        if( shoeState!=oldShoeState )
        {
            oldShoeState = shoeState;
            if( shoeState == SHOE_DOWN ) //todo: check to see how long ago, to prevent double/triple firing
            {
                splash();
                delay(100);
            }
        }
        twinkle();
    }
        
}



void splash()
{
    for( uint8_t i=0; i<19; i++ )
    {
        setLed(i, 6);//why 5?
        delay(15);
    }
    for( uint8_t i=0; i<19; i++ )
    {
        setLed(i, 0);
        //leds[i] = 0;
        delay(15);
    }

/*    
    // fade just isn't visible, as its too fast
    for( uint8_t i=0; i<19; i++ )
    {
        for(uint8_t b=1; b<=5; b++)
        {
            leds[i] = b;
            delay(50);
        }
    }
    for( uint8_t i=0; i<19; i++ )
    {
        for(uint8_t b=1; b<=5; b++)
        {
            leds[i] = 5-b;
            delay(50);
        }
    }
    /* */
}



void twinkle()
{
    static uint8_t index = 4;
    static uint8_t step = 0;
    //if( shoeState==SHOE_DOWN)
    if( step>30 )// 3 steps full on, 15 steps fade to zero, 17 steps off
    {
        index += 7 + ADCH/64*2;//(ADCH>>6);// 7 + (0..3)*2 // Use ADCs least 2 bit to get some random data
        //index = ADCH/12;
        if( index>18 )
            index-=19;
        step = 0;
    }
    int8_t b = 18-step;
    if( b>16 )
        b=16;
    else if( b<0 )
        b=0;
    step++;
        
    setLed(index, b);
    delay(5);

/*
    // 20ms on 80ms off
    setLed(index, 16);
    delay(10);
    for(uint8_t b=0; b<=16; b++ )
    {
        setLed(index, 16-b);
        //leds[index] = 5-b;
        delay(4);//delay(1);
    }
    delay(80);
*/    
}



void setLed(uint8_t index, uint8_t value) // index: 0..18; value 0..16
{
    if( index<19 && value<=16 )
        leds[index] = pgm_read_byte_near(GAMMA+value);
}



void init()
{
    // MCU is awake on power up, then cycles between sleep and awake when pressing the (reset) button
    if( bit_is_set(MCUSR, EXTRF) )// reset button was pressed
    {
        MCUSR = 0;
        if( EEARL==7 )//MCU was awake before the reset
        {
            EEARL = 0;//MCU is asleep
            sleep();
        }
        //else MCU was asleep before reset
    }
    EEARL = 7;//MCU is awake

    
    // Timer 0
    TCCR0B = _BV(CS01) | _BV(WGM02);// /8 prescaler
    TCCR0A = _BV(WGM00)|_BV(WGM01);// fast pwm mode
    OCR0A = 30-1; // 9600000/(8*19*16*30) = 131hz refresh rate  (9.6MHz; /8 scaler; 19 leds; 16 grayscale values)
                  // 9600000/(8*30) = 40khz  40 ISR requests per ms
    TIMSK0 = _BV(TOIE0);//overflow interrupt
    //TCNT0=0;
    

    // ADC
    ADMUX = _BV(ADLAR) | _BV(MUX1); // Vcc as reference voltage; left adjust result; Analog channel ADC2 (PB4)
    ADCSRA = _BV(ADEN) | _BV(ADIE) | _BV(ADPS2) | _BV(ADPS1); // enable adc; enable interrupt; prescaler /64 (150kHz, so 11ksamples/s; so less than 0.1ms per sample)
    
    sei();
}



void sleep()
{
    //power_all_disable();
    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_enable();
    //sleep_bod_disable();
    // go to sleep now
    sleep_cpu();

    // It only wakes up due to a reset, so it never reaches here 
}



void delay(uint16_t ms)
{
    uint16_t currentMs = 0;
    uint16_t startMillis = 0;

    ATOMIC_BLOCK(ATOMIC_FORCEON)
    {
        startMillis = _millis;
    }
    
    do
    {
        ATOMIC_BLOCK(ATOMIC_FORCEON)
        {
            currentMs = _millis;
        }
    }
    while( (currentMs-startMillis) < ms );
}



uint16_t millis()
{
    ATOMIC_BLOCK(ATOMIC_FORCEON)
    {
        return _millis;
    }    
}



EMPTY_INTERRUPT(INT0_vect);



ISR(ADC_vect)
{   
    //called every 64ms
    const uint8_t highThreshold = 7*12;
    const uint8_t lowThreshold = 4*12;
    const uint8_t minCount = 3;
    static uint8_t counter = 0;
  
    uint8_t value = ADCH; // get 8 most significant bits from the ADC
   
    if( shoeState == SHOE_UP )
    {
        if( value < lowThreshold )
        {
            counter++;
            if( counter>minCount )
            {
                shoeState = SHOE_DOWN;
                counter = 0;
            }
        }
        else
        {
            counter = 0;
        }
    }
    else //if( shoeState == SHOE_DOWN )
    {
        if( value > highThreshold )
        {
            counter++;
            if( counter>minCount )
            {
                shoeState=SHOE_UP;// change state
                counter = 0;
            }
        }
        else
        {
            counter = 0;
        }
    }
}



ISR(TIM0_OVF_vect) // 91 instructions; 156 clock cycles
{
    // Most of the time the ISR takes less than 8µs of the 25µs (40kHz)
    static uint8_t ledIndex = 0;
    static uint8_t brightnessStep = 0;
    static uint8_t msCounter = 0;
    static uint8_t bufferPORTB = 0;
    static uint8_t bufferDDRB = 0;
    static uint8_t doMeasurement = 0;
    static uint16_t cycleCount = 0;

    if( doMeasurement )
    {
        // turn on IR led
        DDRB = pgm_read_byte_near(ledDDRs+19);
        PORTB = pgm_read_byte_near(ledPORTs+19);
        doMeasurement--;
        if( !doMeasurement )// have the IR led on for 20 cycles before doing the measurement
            ADCSRA |= _BV(ADSC);
    }
    else
    {
        DDRB = bufferDDRB;
        PORTB = bufferPORTB;

        // find next index and brightnesStep that is nonezero
        bool done = false;
        while( ledIndex<19 && !done) // this will occasionally make the ISR take more than 25µs (40kHz) to find the next leds, or reach the end of the index
        {
            if( brightnessStep < leds[ledIndex] )
                done = true;
            else
            {
                ledIndex++;
                brightnessStep = 0;
            }
        }
/*        
        while( brightnessStep<16 && !done )
        {
            while( ledIndex<19 && !done)
            {
                if( brightnessStep < leds[ledIndex] )
                    done = true;
                else
                    ledIndex++;
            }
            if( !done )
            {
                ledIndex = 0;
                brightnessStep++;
            }
        }
*/        
        if( done )
        {
            bufferDDRB = pgm_read_byte_near(ledDDRs+ledIndex);
            bufferPORTB = pgm_read_byte_near(ledPORTs+ledIndex);

            // next brightness for current led; if brightness at max, go to next led
            brightnessStep++;
            if( brightnessStep==16 )
            {
                brightnessStep = 0;
                ledIndex++;
            }
/*            
            if( ledIndex==19 )//todo: 19
            {
                ledIndex=0;
                brightnessStep++;
                if( brightnessStep==16 )
                    brightnessStep = 0;//NOOOOOO
            }
*/            
            
            cycleCount++;
        }
        else
        {
            bufferPORTB = 0;
            bufferDDRB = 0;
            if( cycleCount<16 )
            {
                // skip cycles till the minimum number of 16 cycles has been reached
                cycleCount++;
            }
            else
            {
                // restart from the beginning
                ledIndex = 0;
                brightnessStep = 0;
                cycleCount = 0;
            }
        }
        //if( brightnessStep < (1<<(leds[ledIndex]))/2 )
/*        
        if( brightnessStep < leds[ledIndex] )
        {
            bufferDDRB = pgm_read_byte_near(ledDDRs+ledIndex);
            bufferPORTB = pgm_read_byte_near(ledPORTs+ledIndex);
        }
        else
        {
            bufferPORTB = 0;
            bufferDDRB = 0;
        }


        ledIndex++;
        if( ledIndex==20 )//todo: 19
        {
            ledIndex=0;
            brightnessStep++;
            if( brightnessStep==16 )
                brightnessStep = 0;
        }
*/        
    }


    msCounter++;
    if( msCounter>=40 )
    {
        _millis++;
        msCounter=0;

        if( (_millis % 64)==0 )// do measurement every 64ms; ~16 times a second (verified on logic analyzer)
            doMeasurement = 20;
    }

}



/*
 * old ISR for the ADC. which is too big, and would accidentily trigger, if the ADC values where slightly erratic.
ISR(ADC_vect) //216 bytes of flash
{   //called every 64ms
    wdt_reset();
    static uint8_t highValue = 127;
    static uint8_t lowValue = 127-40;
    static uint8_t value = 127-40;
    static uint8_t stateDuration = 0;
  
    //value = ADCH; // get 8 most significant bits from the ADC
    value = ((int)value*4+ADCH)/5; // get a rolling average
    uint8_t half = highValue/2 + lowValue/2;


//    // if shoe is stuck in a state for to long (difference enviroment/floor), either lower the highValue, or raise the lowValue
//    stateDuration++;
//    if( (stateDuration > 15000/64) && ( (highValue-lowValue)>40 ) )// no state change for at least 15 seconds; and difference between highValue and lowValue is large enough
//    {
//        if( shoeState==SHOE_UP )
//            lowValue = half;
//        else
//            highValue = half;
//    }
    
    //leds[3] = (value>highValue)?5:0;
    
    if( shoeState == SHOE_UP )
    {
        if( value < half )
        {
            shoeState=SHOE_LOWERING;// change state
            lowValue = value;// reset lowValue
            stateDuration = 0;
        }
        if( value > highValue )
            highValue = value;
    }
    else if( shoeState == SHOE_DOWN )
    {
        if( value > half )
        {
            shoeState=SHOE_UP;// change state
            highValue = value;// reset highValue (this will also prevent hysteresis)
            stateDuration = 0;
        }
        if( value < lowValue )
            lowValue = value;
    }
    else // shoeState ==  SHOE_LOWERING
    {
        if( value < lowValue ) //todo: <=
        {
            lowValue = value;
        }
        else
        {
            shoeState = SHOE_DOWN;
            //stateDuration = 0;//not neccessarily needed
        }
    }    
}
*/
